@@ -1,6 +1,11 @@
 """
 This Module will serve as an alternative and extension to opencv - hence the name
 
+Some notes re compatibility with OpenCV.
+
+ 1. Sizes: OpenCV expects sizes to be passed as [W x H]. However, in following with normal matrix
+    manipulations, Images are passed as [H x W x 3]
+
 This program is free software: you can redistribute it and/or modify it under the terms of the GNU
 General Public License as published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.
@@ -22,6 +27,7 @@ from threading import Thread
 import numpy as np
 import time as tm
 import cv2
+import os
 
 
 # Define Some Constants
@@ -29,6 +35,20 @@ VP_CUR_PROP_POS_MSEC = -100  # Current frame (rather than next one) in MS
 VP_CUR_PROP_POS_FRAMES = -101  # Current frame (rather than next one) in index
 
 
+class FourCC:
+    """
+    FourCC Wrapper to allow conversions between Integer and String Representation
+    """
+
+    @classmethod
+    def to_int(cls, fourcc) -> int:
+        return int(cv2.VideoWriter_fourcc(*fourcc))
+
+    @classmethod
+    def to_str(cls, fourcc) -> str:
+        return "".join([chr((int(fourcc) >> 8 * i) & 0xFF) for i in range(4)])
+
+
 class Homography:
     """
     Class for fitting a Homography:
@@ -39,9 +59,10 @@ class Homography:
         """
         Initialiser
 
-        :param image_coords: A 2D Numpy array of image coordinates, with x/y along the second axis. Must be of length at
-                                least 4.
-        :param world_coords: A 2D Numpy arra of corresponding world-coordinates: must be same shape as image_coords
+        :param image_coords: A 2D Numpy array of image coordinates, with x/y along the second axis.
+                             Must be of length at least 4.
+        :param world_coords: A 2D Numpy arra of corresponding world-coordinates: must be same shape
+                            as image_coords
         """
         self.toImg = cv2.findHomography(world_coords, image_coords)[0]
         self.toWrld = cv2.findHomography(image_coords, world_coords)[0]
@@ -50,8 +71,8 @@ class Homography:
         """
         Convert world coordinates to image coordinates
 
-        :param points: 2D Numpy array, with the last dimension of size 2 (X/Y coordinates): if 1D will be automatically
-                       promoted to 2D
+        :param points: 2D Numpy array, with the last dimension of size 2 (X/Y coordinates): if 1D
+                       will be automatically promoted to 2D
         :return:    Image Coordinates
         """
         return np.squeeze(
@@ -64,8 +85,8 @@ class Homography:
         """
         Convert Image coordinates to world-coordinates
 
-        :param points: 2D Numpy array, with the last dimension of size 2 (X/Y coordinates): if 1D will be automatically
-                       promoted to 2D
+        :param points: 2D Numpy array, with the last dimension of size 2 (X/Y coordinates): if 1D
+                       will be automatically promoted to 2D
         :return:    World Coordinates
         """
         return np.squeeze(
@@ -75,17 +96,201 @@ class Homography:
         )
 
 
-def expand_box(c, size):
+class BoundingBox:
     """
-    Create a Rectangle of the specified size, centred at the point. Note that since this is aimed for images, it assumes
-    that Y grows downwards (this is relevant in specifying what is meant by the top-left corner)
+    A Wrapper for describing Bounding Boxes
 
-    :param c:       Centre (2-tuple/array, X/Y)
-    :param size:    The size of the rectangle (2-tuple/array, width/height)
-    :return:        Four corners of the bounding box, clockwise, from top-left corner
+    The class supports both accessing through properties as well as indexing style access []. Note
+    that the Class deals with floating point values and as such, care must be taken when handling
+    center coordinates to convert to correct pixel indices.  Note also that it assumes that image
+    coordinates grow downwards and to the right, meaning that BR > TL always. No checks are done
+    for this.
     """
-    x, y = np.asarray(size) / 2
-    return np.asarray(((c - (x, y)), (c + (x, -y)), (c + (x, y)), (c + (-x, y))))
+
+    def __init__(self, tl=None, br=None, sz=None, c=None):
+        """
+        Initialiser.
+
+        Two and only two of the parameters must be defined
+        :param tl:  Top-Left Corner (X/Y)
+        :param br:  Bottom-Right Corner (X/Y)
+        :param sz:  Size (X/Y)
+        :param c:   Center (X/Y)
+        """
+        if sum(1 for l in locals().values() if l is not None) != 3:
+            raise ValueError("Exactly two of tl/br/c/sz must be specified!")
+
+        self.TL = np.asarray(tl, dtype=float) if tl is not None else None
+        self.BR = np.asarray(br, dtype=float) if br is not None else None
+        self.C = np.asarray(c, dtype=float) if c is not None else None
+        self.SZ = np.asarray(sz, dtype=float) if sz is not None else None
+
+    def __repr__(self):
+        return (
+            f"[({self.top_left[0]:.1f}, {self.top_left[1]:.1f}), "
+            f"({self.bottom_right[0]:.1f}, {self.bottom_right[1]:.1f})]"
+        )
+
+    def __eq__(self, other):
+        """
+        Equality is defined in terms of top-left and bottom-right values. Note that this is
+        susceptible to the usual issue in comparing equality of floating point values!
+        """
+        return np.array_equal(self.top_left, other.top_left) and np.array_equal(
+            self.bottom_right, other.bottom_right
+        )
+
+    @property
+    def top_left(self):
+        if self.TL is None:
+            if self.BR is not None:
+                if self.SZ is not None:
+                    self.TL = self.BR - self.SZ
+                else:
+                    self.TL = self.C - (self.BR - self.C)
+            else:
+                self.TL = self.C - self.SZ / 2
+        return self.TL
+
+    @property
+    def bottom_right(self):
+        if self.BR is None:
+            if self.TL is not None:
+                if self.SZ is not None:
+                    self.BR = self.TL + self.SZ
+                else:
+                    self.BR = self.C + (self.C - self.TL)
+            else:
+                self.BR = self.C + self.SZ / 2
+        return self.BR
+
+    @property
+    def size(self):
+        if self.SZ is None:
+            if self.TL is not None:
+                if self.BR is not None:
+                    self.SZ = self.BR - self.TL
+                else:
+                    self.SZ = (self.C - self.TL) * 2
+            else:
+                self.SZ = (self.BR - self.C) * 2
+        return self.SZ
+
+    @property
+    def center(self):
+        if self.C is None:
+            if self.TL is not None:
+                if self.BR is not None:
+                    self.C = (self.TL + self.BR) / 2
+                else:
+                    self.C = self.TL + self.SZ / 2
+            else:
+                self.C = self.BR - self.SZ / 2
+        return self.C
+
+    def __getitem__(self, item):
+        item = item.lower()
+        if item == "tl":
+            return self.top_left
+        elif item == "br":
+            return self.bottom_right
+        elif item == "c":
+            return self.center
+        elif item == "sz":
+            return self.size
+        else:
+            raise ValueError("Invalid Attribute")
+
+    def corners(self):
+        """
+        Returns all corners, in a clockwise fashion, starting from top-left
+        :return:
+        """
+        x, y = self.size / 2
+        c = self.C
+        return np.asarray(((c - (x, y)), (c + (x, -y)), (c + (x, y)), (c + (-x, y))))
+
+    def iou(self, other):
+        pass
+
+
+def line(img, pt1, pt2, color, thickness=1, lineType=8, shift=0, linestyle="-"):
+    """
+    Draws a line (much like cv2.line()) but allows for other line-styles.
+
+    Args:
+        linestyle: specifies the line-style: currently, supports:
+            '-': Standard (default)
+            '--': Dashed Line
+            '.': Dotted line
+        For other arguments, see cv2.line(). Note that lineType and shift are ignored if
+        linestyle is not '-'
+
+    @TODO: Some fancy checks to ensure that line ends with a dash.
+    """
+
+    if linestyle == "-":
+        return cv2.line(
+            img,
+            (int(pt1[0]), int(pt1[1])),
+            (int(pt2[0]), int(pt2[1])),
+            color,
+            thickness,
+            lineType,
+            shift,
+        )
+
+    elif linestyle == "--":
+        pt1 = np.asarray(pt1)
+        pt2 = np.asarray(pt2)
+
+        # Some Calculations
+        seg_len = 12 * thickness
+        line_length = np.sqrt((np.square(pt2 - pt1)).sum())
+        dvect_sld = np.divide(pt2 - pt1, line_length) * 7 * thickness
+        dvect_full = np.divide(pt2 - pt1, line_length) * seg_len
+
+        # Draw
+        for i in range(np.ceil(line_length / seg_len).astype(int)):
+            st = pt1 + dvect_full * i
+            nd = st + dvect_sld
+            cv2.line(
+                img,
+                (int(st[0]), int(st[1])),
+                (int(min(nd[0], pt2[0])), int(min(nd[1], pt2[1]))),
+                color,
+                thickness,
+                cv2.LINE_AA,
+            )
+
+    elif linestyle == ".":
+        pt1 = np.asarray(pt1)
+        pt2 = np.asarray(pt2)
+
+        # Some Calculations
+        seg_len = 4 * thickness
+        line_length = np.sqrt((np.square(pt2 - pt1)).sum())
+        dvect_full = np.divide(pt2 - pt1, line_length) * seg_len
+
+        for i in range(np.ceil(line_length / seg_len).astype(int)):
+            st = pt1 + dvect_full * i
+            cv2.circle(img, (int(st[0]), int(st[1])), thickness, color, -1, cv2.LINE_AA)
+        cv2.circle(img, (int(pt2[0]), int(pt2[1])), thickness, color, -1, cv2.LINE_AA)
+
+
+def rectangle(img, pt1, pt2, color, thickness=1, lineType=8, shift=0, linestyle="-"):
+    """
+    Draws a Rectangle at the specified position
+
+    Args:
+        linestyle: specifies the line-style (see cvext.line())
+        For other arguments see cv2.rectangle()
+
+    """
+    line(img, pt1, (pt2[0], pt1[1]), color, thickness, lineType, shift, linestyle)
+    line(img, (pt2[0], pt1[1]), pt2, color, thickness, lineType, shift, linestyle)
+    line(img, pt1, (pt1[0], pt2[1]), color, thickness, lineType, shift, linestyle)
+    line(img, (pt1[0], pt2[1]), pt2, color, thickness, lineType, shift, linestyle)
 
 
 def intersection_over_union(ground_truth, prediction):
@@ -109,14 +314,32 @@ def intersection_over_union(ground_truth, prediction):
         return 0
 
     # Compute the union
-    union = (
-        ground_truth[2] * ground_truth[3] + prediction[2] * prediction[3] - intersection
-    )
+    union = ground_truth[2] * ground_truth[3] + prediction[2] * prediction[3] - intersection
 
     # Return Intersection over Union
     return intersection / union
 
 
+class TimeFrame:
+    """
+    A Convertor class for switching between time/frame numbers.
+
+    Note:
+        time is always returned in MS.
+        frames are by default 0-offset, but this can be changed in the initialiser
+    """
+
+    def __init__(self, fps=25, frm_offset=0):
+        self.FPS = fps
+        self.offset = frm_offset
+
+    def to_frame(self, ms):
+        return int(np.round(ms * self.FPS / 1000) + self.offset)
+
+    def to_time(self, frm):
+        return (frm - self.offset) * 1000 / self.FPS
+
+
 class VideoParser:
     """
     The Video Parser (Wrapper) Object
@@ -151,9 +374,11 @@ class VideoParser:
     def start(self, start=None, stride=1):
         """
         Start the Parsing Loop
-        :param start:   If not None (default) then signifies an index of the frame at which to start
-        :param stride:  How much to stride: default is to just add 1... (this is in terms of frames). Note that when
-                        striding, the last frame is ALWAYS read even if it is not a multiple of the stride...
+        :param start:   If not None (default) then signifies an index of the frame at which to
+        start (0-offset)
+        :param stride:  How much to stride: default is to just add 1 (this is in terms of frames).
+                        Note that when striding, the last frame is ALWAYS read even if it is not
+                        a multiple of the stride.
         :return:        True if successful, false otherwise
         """
         # Check that not already processing
@@ -223,9 +448,7 @@ class VideoParser:
                     self.properties[cv2.CAP_PROP_POS_MSEC]
                 )
                 self.properties[cv2.CAP_PROP_POS_MSEC] = _data[0]
-                self.properties[VP_CUR_PROP_POS_FRAMES] = self.properties[
-                    cv2.CAP_PROP_POS_FRAMES
-                ]
+                self.properties[VP_CUR_PROP_POS_FRAMES] = self.properties[cv2.CAP_PROP_POS_FRAMES]
                 self.properties[cv2.CAP_PROP_POS_FRAMES] = _data[1]
                 return True, _data[2]
             else:
@@ -277,9 +500,7 @@ class VideoParser:
         # Store/Initialise some properties
         self.properties[cv2.CAP_PROP_POS_MSEC] = stream.get(cv2.CAP_PROP_POS_MSEC)
         self.properties[cv2.CAP_PROP_POS_FRAMES] = stream.get(cv2.CAP_PROP_POS_FRAMES)
-        self.properties[cv2.CAP_PROP_FRAME_HEIGHT] = stream.get(
-            cv2.CAP_PROP_FRAME_HEIGHT
-        )
+        self.properties[cv2.CAP_PROP_FRAME_HEIGHT] = stream.get(cv2.CAP_PROP_FRAME_HEIGHT)
         self.properties[cv2.CAP_PROP_FRAME_WIDTH] = stream.get(cv2.CAP_PROP_FRAME_WIDTH)
         self.properties[cv2.CAP_PROP_FPS] = stream.get(cv2.CAP_PROP_FPS)
         self.properties[cv2.CAP_PROP_FRAME_COUNT] = stream.get(cv2.CAP_PROP_FRAME_COUNT)
@@ -320,29 +541,64 @@ class VideoParser:
         stream.release()
 
 
+class FrameGetter:
+    """
+    Wrapper for reading successive frames from disk.
+    """
+
+    def __init__(self, path, fmt="{:06d}.jpg"):
+        """
+        Initialiser
+
+        :param path: Path to the extracted frames
+        :param fmt:  The format mode for the name
+        """
+        if not os.path.isdir(path):
+            raise RuntimeError(f'Specified Directory "{path}" does not exist!')
+        self.Path = path
+        self.Fmt = fmt
+
+    def __getitem__(self, item):
+        """
+        Retrieve a frame by Number
+
+        :param item: Frame Number. Note that this is translated directly to image name.
+        :return: OpenCV Image
+        :raises  ValueError if the frame does not exist
+        """
+        _pth = os.path.join(self.Path, self.Fmt.format(item))
+        if os.path.exists(_pth):
+            return cv2.imread(_pth)
+        raise ValueError(f"Image {item} does not exist.")
+
+
 class SwCLAHE:
     """
-    Class for implementing the CLAHE algorithm by way of a sliding window. Note that this is a loose adaptation, and I
-    do cut some corners in the interest of some efficiency. Note, that this requires the Image data to be in 8-bit
-    Format! The reason for this implementation is to separate the histogram computation from the equalisation step.
+    Class for implementing the CLAHE algorithm by way of a sliding window. Note that this is a loose
+    adaptation, and I do cut some corners in the interest of some efficiency. Note, that this
+    requires the Image data to be in 8-bit Format! The reason for this implementation is to
+    separate the histogram computation from the equalisation step.
 
     Note that the algorithm is modified as follows to allow a history:
      a) Keep track of per-pixel raw-counts in a histogram.
-     b) For the true lut, perform clipping based on the number of frames used in the histogram computation.
+     b) For the true lut, perform clipping based on the number of frames used in the histogram
+        computation.
     """
 
     def __init__(self, imgSize, clipLimit=2.0, tileGridSize=(8, 8), padding="reflect"):
         """
         Initialiser
 
-        :param imgSize:         The Dimensions of the Image (Width[C] x Height[R])
-        :param clipLimit:       The Clip Limit to Employ. Note that this will be the clip-limit per image added to the
-                                histogram (computed retroactively). If not required, pass None
-        :param tileGridSize:    The Tile-Size to compute with. Note that in our case, this signifies the padding around
-                                the pixel, which is a deviation from the OpenCV Implementation! The padding is in terms
-                                of width and height respectivel.
-        :param padding:         Type of padding to employ when computing along the edges. See the documentation for
-                                numpy.pad
+        :param imgSize:      The Dimensions of the Image (Width[C] x Height[R])
+        :param clipLimit:    The Clip Limit to Employ. Note that this will be the clip-limit per
+                             image added to the histogram (computed retroactively). If not
+                             required, pass None
+        :param tileGridSize: The Tile-Size to compute with. Note that in our case, this signifies
+                             the padding around the pixel, which is a deviation from the OpenCV
+                             Implementation! The padding is in terms of width and height
+                             respectivel.
+        :param padding:      Type of padding to employ when computing along the edges. See the
+                             documentation for numpy.pad
         """
         # Store some Values for later
         self.__W, self.__H = imgSize
@@ -377,10 +633,11 @@ class SwCLAHE:
 
     def update_model(self, img):
         """
-        Update the Histogram and perform Clipping. This creates a usable LUT, and is equivalent to calling
-        update_histogram() followed by generate_lut().
+        Update the Histogram and perform Clipping. This creates a usable LUT, and is equivalent to
+        calling update_histogram() followed by generate_lut().
 
-        :param img: Input image to use to update the Histogram with. Must be a single channel image of type uint8.
+        :param img: Input image to use to update the Histogram with. Must be a single channel image
+                    of type uint8.
         :return:    self, for chaining
         """
         return self.update_histogram(img).generate_lut()
@@ -389,7 +646,8 @@ class SwCLAHE:
         """
         Update the Histogram only, without perform any clipping
 
-        :param img: Input image to use to update the Histogram with. Must be a single channel image of type uint8
+        :param img: Input image to use to update the Histogram with. Must be a single channel image
+                    of type uint8
         :return:    self, for chaining
         """
         # First PAD the image: this will allow computation being much easier...
@@ -434,8 +692,8 @@ class SwCLAHE:
 
     def apply(self, img, clear=True):
         """
-        Convenience Method for joining together updating of histogram and transform... This emulates the default
-        CLAHE implementation which clears the histogram after each iteration
+        Convenience Method for joining together updating of histogram and transform... This emulates
+        the default CLAHE implementation which clears the histogram after each iteration
 
         :param img:     The image to operate on
         :param clear:   If True, then re-generate the histogram.
@@ -447,8 +705,7 @@ class SwCLAHE:
 
     @staticmethod
     @jit(
-        signature_or_function=(uint8[:, :], uint8, uint8, uint16[:, :, :]),
-        nopython=True,
+        signature_or_function=(uint8[:, :], uint8, uint8, uint16[:, :, :]), nopython=True,
     )
     def __update_hist(padded, row_pad, col_pad, hist):
         """
@@ -471,8 +728,8 @@ class SwCLAHE:
         for r_img in range(*valid_rows):
             # Get the Histogram Row we are working on...
             r_hst = r_img - row_pad
-            # Compute First Pixel:
-            # Not that there is a special case when this is the top-left corner, which we must compute from scratch.
+            # Compute First Pixel: Not that there is a special case when this is the top-left
+            # corner, which we must compute from scratch.
             if r_hst == 0:
                 for nbh_r in range(row_pad * 2 + 1):
                     for nbh_c in range(col_pad * 2 + 1):
@@ -483,7 +740,8 @@ class SwCLAHE:
                 # Compute the Previous/Next Row (in image space)
                 r_prev = r_img - row_pad - 1
                 r_next = r_img + row_pad
-                # Iterate over the columns, subtracting the previous row and adding the next one in turn
+                # Iterate over the columns, subtracting the previous row and adding the next one in
+                # turn
                 for nbh_c in range(col_pad * 2 + 1):
                     hist[r_hst, 0, padded[r_prev, nbh_c]] -= 1
                     hist[r_hst, 0, padded[r_next, nbh_c]] += 1
@@ -501,15 +759,15 @@ class SwCLAHE:
 
     @staticmethod
     @jit(
-        signature_or_function=(double[:, :, ::1], double, uint8[:, :, ::1], double),
-        nopython=True,
+        signature_or_function=(double[:, :, ::1], double, uint8[:, :, ::1], double), nopython=True,
     )
     def __clip_limit(hist, limit, lut, scaler):
         """
         Here hist should be float (to avoid truncation), but lut is integer!
 
         :param hist:   The Original Histogram (raw)
-        :param limit:  This should take into consideration the number of samples seen so far (i.e. a multiplicated)
+        :param limit:  This should take into consideration the number of samples seen so far (i.e.
+                       a multiplicated)
         :param lut:    The Lookup table (placeholder)
         :param scaler: should be 256 / boxsize
         :return:
@@ -527,7 +785,8 @@ class SwCLAHE:
                         hist[r, c, h] = limit
                 # Now Redistribute - Note that I will ignore residuals (handled through rounding)
                 hist[r, c, :] += to_clip / 256
-                # Now Transform to Lookup Table (rounding) - Had to do this manually due to issues with Numba!
+                # Now Transform to Lookup Table (rounding) - Had to do this manually due to issues
+                # with Numba!
                 cumsum = 0
                 for h in range(256):
                     cumsum += hist[r, c, h]
@@ -535,8 +794,7 @@ class SwCLAHE:
 
     @staticmethod
     @jit(
-        signature_or_function=(uint8[:, :, ::1], uint8[:, ::1], uint8[:, ::1]),
-        nopython=True,
+        signature_or_function=(uint8[:, :, ::1], uint8[:, ::1], uint8[:, ::1]), nopython=True,
     )
     def __transform(lut, img, out):
         """
