@@ -113,8 +113,9 @@ def exp10(a):
 
 def clamp(a, cutoff, limits):
     """
-    Clamp the values in a to limits at the cutoff: specifically, values less than or equal to the cutoff are converted
-    to limits[0] and values greather than cutoff are converted to limits[1]. This is a counterpart to the clip function.
+    Clamp the values in a to limits at the cutoff: specifically, values less than or equal to the
+    cutoff are converted to limits[0] and values greather than cutoff are converted to limits[1].
+    This is a counterpart to the clip function.
 
     :param a:       Array
     :param cutoff:  Cutoff value for clamping from
@@ -130,7 +131,8 @@ def clamp(a, cutoff, limits):
 
 def mad(a, b, axis=None, keepdims=False):
     """
-    Compute the Mean Absolute Deviation between two arrays: they must be of same size or broadcastable
+    Compute the Mean Absolute Deviation between two arrays: they must be of same size or
+    broadcastable
 
     :param a:           First Array
     :param b:           Second Array
@@ -159,10 +161,11 @@ def value_map(array, _to, _from=None, shuffle=False):
 
     :param array:  Array to map
     :param _to:    Values to map to
-    :param _from:  Values to map from: if not present, it is assumed that the entries in to are ordered in their domain
-                    [0, len(to)-1]
-    :param shuffle: Optimisation: if True, does not need to use searchsorted (uses just array indexing). Only set this
-                     to True iff _from contains all contiguous values from 0 up to len(_to) - 1 (also no duplicates).
+    :param _from:  Values to map from: if not present, it is assumed that the entries in to are
+                   ordered in their domain [0, len(to)-1]
+    :param shuffle: Optimisation: if True, does not need to use searchsorted (uses just array
+                    indexing). Only set this to True iff _from contains all contiguous values
+                    from 0 up to len(_to) - 1 (also no duplicates).
     :return:       Mapped Array
     """
     # If just shuffling, then use just array indexing...
@@ -187,18 +190,21 @@ def run_lengths(a, how="I", return_values=False, return_positions=False):
     :param a: Array If not one-d, the input is flattened.
     :param how: How to treat values including np.NaN
                     A: use All values (including treating NaN as its own de-facto value)
-                    I: Ignore NaN values but treat all others as before. Note that if an NaN interrupts a stream of
-                       same values, then these will be treated as SEPARATE streams.
+                    I: Ignore NaN values but treat all others as before. Note that if an NaN
+                       interrupts a stream of same values, then these will be treated as SEPARATE
+                       streams.
                     O: compute Only NaNs
     :param return_values: If true, return also the Key (value) for each run.
     :param return_positions: If true, return also the start position of each run.
     :return: lengths, [values], [positions]
-               The first array contains the run lengths: optionally, the next array (if present) is the values for each
-               run, and finally (if present) the position of the start of each run.
+               The first array contains the run lengths: optionally, the next array (if present) is
+               the values for each run, and finally (if present) the position of the start of
+               each run.
     """
     # Prepare
-    #  We fist compute assuming that we are going to use everything! This allows us to have correct position information
-    #  We also need to convert arrays of NaN to lists so that NaN works. This will slow down but cannot be helped.
+    #  We fist compute assuming that we are going to use everything! This allows us to have correct
+    #  position information We also need to convert arrays of NaN to lists so that NaN works.
+    #  This will slow down but cannot be helped.
     if type(a) == np.ndarray:
         a = [i if not np.isnan(i) else np.nan for i in a.flatten()]
     rls = np.asarray(
@@ -213,15 +219,11 @@ def run_lengths(a, how="I", return_values=False, return_positions=False):
     elif how.lower() == "i":
         n_nan = ~np.isnan(rls[:, 1])
         lens, vals, pos = (
-            (rls[n_nan, 0], rls[n_nan, 1], pos[n_nan])
-            if len(n_nan) > 0
-            else ([], [], [])
+            (rls[n_nan, 0], rls[n_nan, 1], pos[n_nan]) if len(n_nan) > 0 else ([], [], [])
         )
     elif how.lower() == "o":
         nan = np.isnan(rls[:, 1])
-        lens, vals, pos = (
-            (rls[nan, 0], rls[nan, 1], pos[nan]) if len(nan) > 0 else ([], [], [])
-        )
+        lens, vals, pos = (rls[nan, 0], rls[nan, 1], pos[nan]) if len(nan) > 0 else ([], [], [])
     else:
         raise ValueError('Incorrect format for "how" parameter.')
     # Now Return
@@ -235,10 +237,12 @@ def run_lengths(a, how="I", return_values=False, return_positions=False):
 
 def array_nan_equal(left, right):
     """
-    Compares two numpy arrays to test for equality, but considers np.nan values to be equal to each other
+    Compares two numpy arrays to test for equality, but considers np.nan values to be equal to each
+    other
     :param left: The Left array to consider
     :param right: The Right array to consider
-    :return: True if left is the same shape as right, and all elements are equal (including nan's at same positions)
+    :return: True if left is the same shape as right, and all elements are equal (including nan's at
+             same positions)
     """
     # Ensure that Arrays (potentially convert from list)
     try:
@@ -294,16 +298,15 @@ def non_diag(a: np.ndarray):
 
 def make_diagonal(on_diag, off_diag, size):
     """
-    Create a Diagonally-dominant matrix with on_diag elements on the main diagonal, and off_diag elsewhere
+    Create a Diagonally-dominant matrix with on_diag elements on the main diagonal, and off_diag
+    elsewhere
 
     :param on_diag:     Scalar: Value to put on diagonal
     :param off_diag:    Scalar: Value to put off the diagonal
     :param size:        Scalar: Size of the matrix
     :return:            Diagonal matrix
     """
-    return np.eye(size) * on_diag + non_diag(
-        np.full(shape=[size, size], fill_value=off_diag)
-    )
+    return np.eye(size) * on_diag + non_diag(np.full(shape=[size, size], fill_value=off_diag))
 
 
 def maximise_trace(x):
@@ -365,10 +368,12 @@ class Dirichlet:
         """
         Initialise the Dirichlet with a particular alpha
 
-        :param alpha: Alpha (Concentration) parameter. This is a K-D Matrix, where the last dimension is the probability
-        :param ignore_zeros: If true, will ignore dimensions which are expected to be 0. This is inferred from the alpha
-                             matrix, specifically ignoring dimensions which are 1. This is based on the assumption that
-                             such dimensions will always be zero and will be known in advance.
+        :param alpha: Alpha (Concentration) parameter. This is a K-D Matrix, where the last
+                      dimension is the probability
+        :param ignore_zeros: If true, will ignore dimensions which are expected to be 0. This is
+                             inferred from the alpha matrix, specifically ignoring dimensions
+                             which are 1. This is based on the assumption that such dimensions
+                             will always be zero and will be known in advance.
         """
         alpha = np.asarray(alpha)  # K   Dimensional
         # Do some Checks
@@ -386,8 +391,8 @@ class Dirichlet:
 
     def pdf(self, x):
         """
-        Compute the PDF of the passed Matrix (must be of the same dimensionality and size as Alpha). Note that if 0's
-        are passed, then these must align with the alpha's ones.
+        Compute the PDF of the passed Matrix (must be of the same dimensionality and size as Alpha).
+        Note that if 0's are passed, then these must align with the alpha's ones.
 
         :param x: Probability Matrix
         :return: Dirichlet Probability on x
@@ -397,15 +402,16 @@ class Dirichlet:
             raise ValueError("0-valued probabilities are only allowed where alpha=1")
 
         # Note that we can do the below because:
-        #   a) If we are not ignoring zeros, then anything raised to 0 is 1 in any case, and since we are taking product
-        #       this does not matter
-        #   b) If we are ignoring zeros, this is masked by definition, which means that the result of the power is
-        #       masked and so is the product!
+        #   a) If we are not ignoring zeros, then anything raised to 0 is 1 in any case, and since
+        #      we are taking product this does not matter
+        #   b) If we are ignoring zeros, this is masked by definition, which means that the result
+        #      of the power is masked and so is the product!
         return np.prod(np.power(x, self._alpha), axis=-1) / self.norm
 
     def logpdf(self, x):
         """
-        Compute the Log-PDF of the passed Matrix (must be of the same dimensionality and size as Alpha)
+        Compute the Log-PDF of the passed Matrix (must be of the same dimensionality and size as
+        Alpha)
 
         :param x: Probability Matrix
         :return:  Log-Probability on X
@@ -413,17 +419,20 @@ class Dirichlet:
         # Ensure that it is non-zero where zeros are not allowed
         if np.any(x[~self.__zeros] <= 0):
             raise ValueError("0-valued probabilities are only allowed where alpha=1")
-        # In this case, we must mask to avoid the issues with log since -inf * 0 = NaN... however, we also set the error
-        #   state appropriately since the mask does not appear to work on log...
-        # Note that while we may wish to support alpha=1 even when non-zero, in any case, this will not have any effect
-        #   since for alpha=1, _alpha=0 and hence, the multiplication below will be with 0!
+        # In this case, we must mask to avoid the issues with log since -inf * 0 = NaN... however,
+        #   we also set the error state appropriately since the mask does not appear to work on
+        #   log...
+        # Note that while we may wish to support alpha=1 even when non-zero, in any case,
+        #   this will not have any effect since for alpha=1, _alpha=0 and hence,
+        #   the multiplication below will be with 0!
         with np.errstate(divide="ignore"):
             x = np.ma.array(x, mask=self.__zeros)
             return np.sum(np.multiply(np.log(x), self._alpha), axis=-1) - self.lognorm
 
     def logsumpdf(self, x):
         """
-        Compute the Log-PDF of the passed Matrix (must be of the same dimensionality and size as Alpha) and sum it.
+        Compute the Log-PDF of the passed Matrix (must be of the same dimensionality and size as
+        Alpha) and sum it.
 
         :param x: Probability Matrix
         :return:  Log-Sum-Probability on X
@@ -458,8 +467,8 @@ def sum_to_one(x, axis=None, norm=False):
     The function avoids division by-zero errors.
 
     :param x:       Array to normalise
-    :param axis:    If not None (default) specifies axis to normalise across: otherwise, normalisation happens across
-                    flattened array.
+    :param axis:    If not None (default) specifies axis to normalise across: otherwise,
+                    normalisation happens across flattened array.
     :param norm:    If Norm, return also the normaliser
     :return: Normalised Array
     """
@@ -470,11 +479,13 @@ def sum_to_one(x, axis=None, norm=False):
 
 def invert_softmax(x, enforce_unique=None):
     """
-    Return a vector which would yield x under softmax: note that uniqueness is achieved in a number of ways.
+    Return a vector which would yield x under softmax: note that uniqueness is achieved in a number
+    of ways.
 
     :param x: Vector/Matrix of probabilities (last dimension must sum to 1)
-    :param enforce_unique: If None (default) then uniqueness is achieved by enforcing that the inverted space sums to 0.
-                           Otherwise, it specifies an index which is enforced to be 0.
+    :param enforce_unique: If None (default) then uniqueness is achieved by enforcing that the
+                            inverted space sums to 0. Otherwise, it specifies an index which is
+                            enforced to be 0.
     :return:  Inverse Softmax
     """
     # Compute Log(X)
@@ -493,8 +504,8 @@ def invert_softmax(x, enforce_unique=None):
 
 def conditional_entropy(emission, prior=None, base=None):
     """
-    Compute the Conditional Entropy of a Joint Distribution over latent and conditioned (visible) variables.
-    This only supports 2D Emissions (ie one latent variable and one visible)
+    Compute the Conditional Entropy of a Joint Distribution over latent and conditioned (visible)
+    variables. This only supports 2D Emissions (ie one latent variable and one visible)
 
     :param emission:    The L by V conditional probabilities (Latent along the first index)
     :param prior:       The Distribution over L (if None, use vector of equi-probability)
@@ -519,10 +530,12 @@ def conditional_entropy(emission, prior=None, base=None):
 
 def conditional_entropy_2D(emission, prior=None, base=None):
     """
-    Compute the Conditional Entropy of a joint distribution over latent states and conditioned variables
-    This version is designed to be used when the conditioning variables (latent) are 2D: NOT when the visible ones are 2D
+    Compute the Conditional Entropy of a joint distribution over latent states and conditioned
+    variables This version is designed to be used when the conditioning variables (latent) are
+    2D: NOT when the visible ones are 2D
 
-    :param emission: The L1 by L2 by V conditional probabilities (3D with latents along first two indices)
+    :param emission: The L1 by L2 by V conditional probabilities (3D with latents along first two
+                     indices)
     :param prior:    The distribution over L1/L2: if None, will initialise to uniform probability
     :param base:     The numeric base to operate with.
     :return:         Conditional Entropy (scalar)
@@ -542,35 +555,38 @@ def conditional_entropy_2D(emission, prior=None, base=None):
 
     else:
         raise ValueError(
-            "Function does not support Tensors of dimensionality {}. For standard conditional entropy, use"
-            " conditional_entropy().".format(emission.ndim)
+            "Function does not support Tensors of dimensionality {}. For standard conditional "
+            " entropy, use conditional_entropy().".format(emission.ndim)
         )
 
 
 def mutual_information(prior, emission, base=None):
     """
-    Compute the Mutual information between an input (Z) and set of output (X) variables, under the assumption that when
-      there is more than 1 X variable, they are conditionally independent of each other given Z (i.e. the Naive Bayes
-      assumption). Be careful however, that as the number of X variables increases, the dimensionality of the problem
-      explodes (since we need to work with the full joint marginal)!
+    Compute the Mutual information between an input (Z) and set of output (X) variables, under the
+      assumption that when there is more than 1 X variable, they are conditionally independent of
+      each other given Z (i.e. the Naive Bayes assumption). Be careful however, that as the
+      number of X variables increases, the dimensionality of the problem explodes (since we need
+      to work with the full joint marginal)!
 
-    Note (1) - This is NOT the Conditional Mutual Information, which is something different: this is just the MI between
-    a Z and a multi-dimensional X (taken as one variable).
+    Note (1) - This is NOT the Conditional Mutual Information, which is something different: this is
+      just the MI between a Z and a multi-dimensional X (taken as one variable).
 
-    Note (2) - This only supports a 1D latent variable (Z): i.e. there may be multiple (X) variables but only 1 (Z).
+    Note (2) - This only supports a 1D latent variable (Z): i.e. there may be multiple (X) variables
+      but only 1 (Z).
 
     :param prior:       Prior Distribution over Z [1D array]
     :param emission:    Conditional Distribution over X given Z. This can be either:
                             a) 2D Numpy array, with Z along the rows, for one conditional
-                            b) List of 2D Numpy arrays, each constituting a 2D Numpy array (Z along rows) showing the
-                               emission of a variable.
+                            b) List of 2D Numpy arrays, each constituting a 2D Numpy array (Z along
+                               rows) showing the emission of a variable.
     :param base:        The numeric base to operate with.
     :return:            Mutual Information
     """
-    # First Collapse all emissions into 1 by computing outer product along X-axis (i.e. finding the cross-product)
-    #  This is done by a smart use of the self-looping and the outer product function. Basically, we need to find the
-    #  cross-combination of all emissions. To do this, for each latent state (Z=z), we find the probability of each
-    #  combination of Xs by repeatedly (over each variable) finding the outer product between the current flattened
+    # First Collapse all emissions into 1 by computing outer product along X-axis (i.e. finding the
+    #  cross-product) This is done by a smart use of the self-looping and the outer product
+    #  function. Basically, we need to find the cross-combination of all emissions. To do this,
+    #  for each latent state (Z=z), we find the probability of each combination of Xs by
+    #  repeatedly (over each variable) finding the outer product between the current flattened
     #  representation and the next variable in sequence.
     if type(emission) in (list, tuple):
         pXZ = [1 for _ in prior]
@@ -585,9 +601,7 @@ def mutual_information(prior, emission, base=None):
     pX = np.matmul(prior, emission)
 
     # Now Compute Entropies and return
-    return entropy(pX, base=base) - conditional_entropy(
-        emission=emission, prior=prior, base=base
-    )
+    return entropy(pX, base=base) - conditional_entropy(emission=emission, prior=prior, base=base)
 
 
 def markov_stationary(transition):
@@ -606,11 +620,12 @@ def markov_stationary(transition):
 def switch_rate(a, axis=-1, ratio=False):
     """
     Compute the number/frequency of switching at various levels in a time-series.
-    This is guaranteed to work for 1D/2D arrays but the result is unspecified for other dimensionalities
+    This is guaranteed to work for 1D/2D arrays but the result is unspecified for other
+    dimensionalities
 
     :param a:       Array_like
-    :param axis:    Optional: the axis along which to compute, default is the last axis. Note that this can only be one
-                    axis at most.
+    :param axis:    Optional: the axis along which to compute, default is the last axis. Note that
+                    this can only be one axis at most.
     :param ratio:   Whether to return a ratio or absolute value.
     :return:
     """
